defmodule Nulla.ActivityPub do
  alias Nulla.Models.Actor
  alias Nulla.Models.Activity
  alias Nulla.Models.Note
  alias Nulla.Models.InstanceSettings

  @spec context() :: list()
  defp context do
    [
      "https://www.w3.org/ns/activitystreams",
      "https://w3id.org/security/v1",
      Jason.OrderedObject.new(
        manuallyApprovesFollowers: "as:manuallyApprovesFollowers",
        toot: "http://joinmastodon.org/ns#",
        featured: %{"@id" => "toot:featured", "@type" => "@id"},
        featuredTags: %{"@id" => "toot:featuredTags", "@type" => "@id"},
        alsoKnownAs: %{"@id" => "as:alsoKnownAs", "@type" => "@id"},
        movedTo: %{"@id" => "as:movedTo", "@type" => "@id"},
        schema: "http://schema.org#",
        PropertyValue: "schema:PropertyValue",
        value: "schema:value",
        discoverable: "toot:discoverable",
        suspended: "toot:suspended",
        memorial: "toot:memorial",
        indexable: "toot:indexable",
        attributionDomains: %{"@id" => "toot:attributionDomains", "@type" => "@id"},
        Hashtag: "as:Hashtag",
        vcard: "http://www.w3.org/2006/vcard/ns#"
      )
    ]
  end

  @spec actor(Actor.t()) :: Jason.OrderedObject.t()
  def actor(actor) do
    Jason.OrderedObject.new(
      "@context": context(),
      id: actor.ap_id,
      type: actor.type,
      following: actor.following,
      followers: actor.followers,
      inbox: actor.inbox,
      outbox: actor.outbox,
      featured: actor.featured,
      featuredTags: actor.featuredTags,
      preferredUsername: actor.preferredUsername,
      name: actor.name,
      summary: actor.summary,
      url: actor.url,
      manuallyApprovesFollowers: actor.manuallyApprovesFollowers,
      discoverable: actor.discoverable,
      indexable: actor.indexable,
      published: DateTime.to_iso8601(actor.published),
      memorial: actor.memorial,
      publicKey: actor.publicKey,
      tag: actor.tag,
      attachment: actor.attachment,
      endpoints: actor.endpoints,
      icon: actor.icon,
      image: actor.image,
      "vcard:bday": actor.vcard_bday,
      "vcard:Address": actor.vcard_Address
    )
  end

  @spec note(Note.t()) :: Jason.OrderedObject.t()
  def note(note) do
    attachment =
      case note.media_attachments do
        [] ->
          []

        attachments ->
          [
            attachment:
              Enum.map(attachments, fn att ->
                Jason.OrderedObject.new(
                  type: "Document",
                  mediaType: att.mime_type,
                  url: "https://#{note.actor.domain}/files/#{att.file}"
                )
              end)
          ]
      end

    Jason.OrderedObject.new(
      "@context": [
        "https://www.w3.org/ns/activitystreams",
        Jason.OrderedObject.new(sensitive: "as:sensitive")
      ],
      id: "#{note.actor.ap_id}/notes/#{note.id}",
      type: "Note",
      summary: nil,
      inReplyTo: note.inReplyTo,
      published: note.inserted_at,
      url: note.url,
      attributedTo: note.actor.ap_id,
      to: [
        "https://www.w3.org/ns/activitystreams#Public"
      ],
      cc: [
        "#{note.actor.ap_id}/followers"
      ],
      sensitive: note.sensitive,
      content: note.content,
      contentMap: Jason.OrderedObject.new("#{note.language}": note.content),
      attachment: attachment
    )
  end

  @spec activity(Activity.t()) :: Jason.OrderedObject.t()
  def activity(activity) do
    Jason.OrderedObject.new(
      "@context": "https://www.w3.org/ns/activitystreams",
      id: activity.ap_id,
      type: activity.type,
      actor: activity.actor,
      object: activity.object
    )
  end

  @spec following(Actor.t(), Integer.t()) :: Jason.OrderedObject.t()
  def following(actor, total) do
    Jason.OrderedObject.new(
      "@context": "https://www.w3.org/ns/activitystreams",
      id: "#{actor.ap_id}/following",
      type: "OrderedCollection",
      totalItems: total,
      first: "#{actor.ap_id}/following?page=1"
    )
  end

  @spec following(Actor.t(), Integer.t(), List.t(), Integer.t(), Integer.t()) ::
          Jason.OrderedObject.t()
  def following(actor, total, following_list, page, limit) when is_integer(page) and page > 0 do
    data = [
      "@context": "https://www.w3.org/ns/activitystreams",
      id: "#{actor.ap_id}/following?page=#{page}",
      type: "OrderedCollectionPage",
      totalItems: total,
      next: "#{actor.ap_id}/following?page=#{page + 1}",
      prev: "#{actor.ap_id}/following?page=#{page - 1}",
      partOf: "#{actor.ap_id}/following",
      orderedItems: following_list
    ]

    data =
      if page <= 1 do
        Keyword.delete(data, :prev)
      else
        data
      end

    data =
      if page * limit > total do
        data
        |> Keyword.delete(:next)
        |> Keyword.delete(:prev)
      else
        data
      end

    Jason.OrderedObject.new(data)
  end

  @spec followers(Actor.t(), Integer.t()) :: Jason.OrderedObject.t()
  def followers(actor, total) do
    Jason.OrderedObject.new(
      "@context": "https://www.w3.org/ns/activitystreams",
      id: "#{actor.ap_id}/followers",
      type: "OrderedCollection",
      totalItems: total,
      first: "#{actor.ap_id}/followers?page=1"
    )
  end

  @spec followers(Actor.t(), Integer.t(), List.t(), Integer.t(), Integer.t()) ::
          Jason.OrderedObject.t()
  def followers(actor, total, followers_list, page, limit)
      when is_integer(page) and page > 0 do
    data = [
      "@context": "https://www.w3.org/ns/activitystreams",
      id: "#{actor.ap_id}/followers?page=#{page}",
      type: "OrderedCollectionPage",
      totalItems: total,
      next: "#{actor.ap_id}/followers?page=#{page + 1}",
      prev: "#{actor.ap_id}/followers?page=#{page - 1}",
      partOf: "#{actor.ap_id}/followers",
      orderedItems: followers_list
    ]

    data =
      if page <= 1 do
        Keyword.delete(data, :prev)
      else
        data
      end

    data =
      if page * limit > total do
        data
        |> Keyword.delete(:next)
        |> Keyword.delete(:prev)
      else
        data
      end

    Jason.OrderedObject.new(data)
  end

  @spec webfinger(Actor.t()) :: Jason.OrderedObject.t()
  def webfinger(actor) do
    data = [
      subject: "#{actor.preferredUsername}@#{actor.domain}",
      aliases: [
        actor.url,
        actor.ap_id
      ],
      links: [
        Jason.OrderedObject.new(
          rel: "http://webfinger.net/rel/profile-page",
          type: "text/html",
          href: actor.url
        ),
        Jason.OrderedObject.new(
          rel: "self",
          type: "application/activity+json",
          href: actor.ap_id
        )
      ]
    ]

    data =
      if actor.icon do
        Keyword.update!(data, :links, fn links ->
          links ++
            [
              Jason.OrderedObject.new(
                rel: "http://webfinger.net/rel/avatar",
                type: Map.get(actor.icon, :mediaType),
                href: Map.get(actor.icon, :url)
              )
            ]
        end)
      else
        data
      end

    Jason.OrderedObject.new(data)
  end

  @spec nodeinfo(String.t()) :: Jason.OrderedObject.t()
  def nodeinfo(domain) do
    Jason.OrderedObject.new(
      links: [
        Jason.OrderedObject.new(
          rel: "http://nodeinfo.diaspora.software/ns/schema/2.0",
          href: "https://#{domain}/nodeinfo/2.0"
        )
      ]
    )
  end

  @spec nodeinfo(String.t(), Map.t(), InstanceSettings.t()) :: Jason.OrderedObject.t()
  def nodeinfo(version, users, instance) do
    Jason.OrderedObject.new(
      version: "2.0",
      software:
        Jason.OrderedObject.new(
          name: "nulla",
          version: version
        ),
      protocols: [
        "activitypub"
      ],
      services:
        Jason.OrderedObject.new(
          outbound: [],
          inbound: []
        ),
      usage:
        Jason.OrderedObject.new(
          users:
            Jason.OrderedObject.new(
              total: users.total,
              activeMonth: users.month,
              activeHalfyear: users.halfyear
            )
        ),
      openRegistrations: instance.registration,
      metadata:
        Jason.OrderedObject.new(
          nodeName: instance.name,
          nodeDescription: instance.description
        )
    )
  end

  @spec outbox(Actor.t(), Integer.t()) :: Jason.OrderedObject.t()
  def outbox(actor, total) do
    Jason.OrderedObject.new(
      "@context": "https://www.w3.org/ns/activitystreams",
      id: "#{actor.ap_id}/outbox",
      type: "OrderedCollection",
      totalItems: total,
      first: "#{actor.ap_id}/outbox?page=true",
      last: "#{actor.ap_id}/outbox?min_id=0&page=true"
    )
  end

  @spec outbox(Actor.t(), Integer.t(), Integer.t(), List.t()) :: Jason.OrderedObject.t()
  def outbox(actor, max_id, min_id, items) do
    Jason.OrderedObject.new(
      "@context": [
        "https://www.w3.org/ns/activitystreams",
        Jason.OrderedObject.new(
          sensitive: "as:sensitive",
          Hashtag: "as:Hashtag"
        )
      ],
      id: "#{actor.ap_id}/outbox?page=true",
      type: "OrderedCollectionPage",
      next: "#{actor.ap_id}/outbox?max_id=#{max_id}&page=true",
      prev: "#{actor.ap_id}/outbox?min_id=#{min_id}&page=true",
      partOf: "#{actor.ap_id}/outbox",
      orderedItems: items
    )
  end

  @spec activity_note(Note.t()) :: Jason.OrderedObject.t()
  def activity_note(note) do
    Jason.OrderedObject.new(
      id: "#{note.actor.ap_id}/notes/#{note.id}/activity",
      type: "Create",
      actor: note.actor.ap_id,
      published: note.inserted_at |> DateTime.to_iso8601(),
      to: [
        "https://www.w3.org/ns/activitystreams#Public"
      ],
      object:
        Jason.OrderedObject.new(
          id: "#{note.actor.ap_id}/notes/#{note.id}",
          type: "Note",
          content: note.content,
          published: note.inserted_at |> DateTime.to_iso8601(),
          attributedTo: note.actor.ap_id,
          to: [
            "https://www.w3.org/ns/activitystreams#Public"
          ]
        )
    )
  end

  @spec follow_accept(Activity.t()) :: Jason.OrderedObject.t()
  def follow_accept(activity) do
    Jason.OrderedObject.new(
      "@context": "https://www.w3.org/ns/activitystreams",
      id: activity.ap_id,
      type: activity.type,
      actor: activity.actor,
      object: activity.object
    )
  end
end
