defmodule Nulla.Models.Note do
  use Ecto.Schema
  import Ecto.Changeset
  import Ecto.Query
  alias Nulla.Repo
  alias Nulla.Snowflake
  alias Nulla.Models.Actor
  alias Nulla.Models.MediaAttachment

  @primary_key {:id, :integer, autogenerate: false}
  schema "notes" do
    field :inReplyTo, :string
    field :url, :string

    field :visibility, Ecto.Enum,
      values: [:public, :unlisted, :followers, :private],
      default: :public

    field :sensitive, :boolean, default: false
    field :content, :string
    field :language, :string

    belongs_to :actor, Actor
    has_many :media_attachments, MediaAttachment

    timestamps(type: :utc_datetime)
  end

  @doc false
  def changeset(note, attrs) do
    note
    |> cast(attrs, [:content, :visibility, :sensitive, :language, :inReplyTo, :actor_id])
    |> validate_required([:content, :visibility, :sensitive, :language, :actor_id])
  end

  def create_note(attrs) when is_map(attrs) do
    id = Map.get(attrs, :id, Snowflake.next_id())

    %__MODULE__{}
    |> changeset(attrs)
    |> put_change(:id, id)
    |> Repo.insert()
  end

  def get_note(id), do: Repo.get(__MODULE__, id)

  def get_latest_notes(actor_id, limit \\ 20) do
    from(n in __MODULE__,
      where: n.actor_id == ^actor_id,
      order_by: [desc: n.inserted_at],
      limit: ^limit,
      preload: [:actor, :media_attachments]
    )
    |> Repo.all()
  end

  def get_before_notes(actor_id, max_id, limit \\ 20) do
    from(n in __MODULE__,
      where: n.actor_id == ^actor_id and n.id < ^max_id,
      order_by: [desc: n.inserted_at],
      limit: ^limit,
      preload: [:actor, :media_attachments]
    )
    |> Repo.all()
  end

  def get_total_notes_count(actor_id) do
    from(n in __MODULE__, where: n.actor_id == ^actor_id)
    |> Repo.aggregate(:count, :id)
  end
end
